////////////////////////////////////////////////////////////////////// 
// fugue-cc/CaseControl.cpp 
// (c) 2000-2003 Goncalo Abecasis
// 
// This file is distributed as part of the FUGUE source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile FUGUE.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Tuesday July 06, 2004
// 
 
#include "HaploSet.h"
#include "Parameters.h"
#include "Random.h"
#include "Error.h"
#include "Likelihood.h"
#include "SparseLikelihood.h"
#include "HaploFamily.h"

void ShowHeader();
void ReadParameters(int argc, char ** argv);
void LoadData();
void CheckData();
void AnalyseHaplotypes();

// Haplotyping Routines
void HaplotypeFounders(ClusterInfo & info, IntArray & filter);
void HaplotypeSample();
void HaplotypeCaseControl();

// Auxiliary Routines
void MaximizeLikelihood(ClusterInfo & info, IntArray & filter, bool print = false);

// Pedigree file contents
Pedigree ped;

// Pedigree filters
IntArray caseFilter, controlFilter, nullFilter;
int case_n, control_n;

// Input haplotype data
HaplotypeSets * sets;

// Name for input file
String datfile("merlin.dat");
String pedfile("merlin.ped");
String mapfile("merlin.map");
String prefix("fugue");

// Request approximate solution
int approximate = false;

// Number of restarts using random haplotype frequencies
int restarts = 0;
int permutations = 0;

// Print additional information for incremental builds
bool additional = false;

// Convergence threshold
double tol = 1e-6;

// Display threshold
double minfreq = 0.0001;

// Calculation workhorses
Likelihood likelihood;

// Options for haplotype frequency estimation
bool smoothing = false, fineSmoothing = false;

int main(int argc, char ** argv)
   {
   ShowHeader();

   ReadParameters(argc, argv);
   LoadData();
   CheckData();

   AnalyseHaplotypes();

   return 0;
   }

void ShowHeader()
   {
   printf("FUGUE FOR CASE-CONTROL DATA\n"
          "(c) 2001-2003 Goncalo Abecasis\n");
   }

void ReadParameters(int argc, char ** argv)
   {
   ParameterList pl;
   pl.Add(new StringParameter('d', "Data File", datfile));
   pl.Add(new StringParameter('p', "Pedigree File", pedfile));
   pl.Add(new IntParameter('e', "Random Restarts for EM", restarts));
   pl.Add(new IntParameter('s', "Random Permutations for Sample", permutations));
   pl.Read(argc, argv);
   pl.Status();
   }

void LoadData()
   {
   ped.Prepare(datfile);
   ped.Load(pedfile);
   ped.Trim();
   ped.EstimateFrequencies(FREQ_EQUAL, true);
   ped.LumpAlleles(0.0);
   }

void CheckData()
   {
   if (ped.affectionCount < 1)
      error("Pedigree file does not include affection status information");

   int founders[2] = {0, 0};
   int cases[3] = {0, 0, 0};

   for (int i = 0; i < ped.count; i++)
      {
      founders[ped[i].isFounder()]++;
      cases[ped[i].affections[0]]++;
      }

   printf("The pedigree file includes:\n"
          "%d cases, %d controls, %d individuals of unknown phenotype\n"
          "%d founders, %d non-founders\n\n",
          cases[2], cases[1], cases[0],
          founders[1], founders[0]);

   if (founders[0])
      error("This program requires a pedigree file including only unrelated individuals\n");

   if (cases[0])
      error("This program requires all individuals to be phenotyped\n");

   if (cases[1] == 0 || cases[2] == 0)
      error("This program requires a sample of cases AND controls\n");

   case_n = cases[2];
   control_n = cases[1];
   }

void HaplotypeFounders(ClusterInfo & info, IntArray & filter)
   {
   FamilyHaplos engine;

   engine.markers = info.markerIds;

   sets->AllocateAlleleLabels(info.markers);
   for (int m = 0; m < info.markers; m++)
      sets->SetAlleleLabels(m, Pedigree::GetMarkerInfo(info.markerIds[m])->alleleLabels);

   for (int i = 0; i < ped.count; i++)
      if (filter[i] && engine.Haplotype(ped, ped[i]))
         sets->LoadFromMemory(&engine, info.markers);
   }

void MaximizeLikelihood(ClusterInfo & info, IntArray & filter, bool print)
   {
   sets = new HaplotypeSets;

   info.UpdateAlleleCounts();

   HaplotypeFounders(info, filter);

   ((ClusterInfo &) likelihood) = info;
   likelihood.tol = tol;

   likelihood.Initialize();
   likelihood.EM(sets);

   for (int i = 0; i < restarts; i++)
      likelihood.RandomEM(sets);

   if (print) likelihood.Print(0.001, sets);

   delete sets;
   }

void SampleFilters()
   {
   caseFilter.Zero();
   controlFilter.Zero();

   int N = case_n + control_n;
   int c = case_n;

   do {
      if (globalRandom.Next() < (double) c / N)
         caseFilter[N] = 1, c--;
      else
         controlFilter[N] = 1;
   } while (--N);
   }

void MakeFilters()
   {
   caseFilter.Dimension(ped.count);
   controlFilter.Dimension(ped.count);
   caseFilter.Zero();
   controlFilter.Zero();

   for (int i = 0; i < ped.count; i++)
      if (ped[i].affections[0] == 2)
         caseFilter[i] = 1;
      else
         controlFilter[i] = 1;

   nullFilter.Dimension(ped.count);
   nullFilter.Set(1);
   }

void AnalyseHaplotypes()
   {
   ClusterInfo info(ped.markerCount);
   info.markerIds.SetSequence(0, 1);

   // First haplotype each individual group
   MakeFilters();

   // Combined sample
   printf("Haplotyping Combined Sample\n"
          "===========================\n\n");
   MaximizeLikelihood(info, nullFilter, true);
   double lkCombined = likelihood.best_llk;

   // Cases
   printf("Haplotyping Case Sample\n"
          "=======================\n\n");
   MaximizeLikelihood(info, caseFilter, true);
   double lkCases = likelihood.best_llk;

   // Controls
   printf("Haplotyping Control Sample\n"
          "==========================\n\n");
   MaximizeLikelihood(info, controlFilter, true);
   double lkControls = likelihood.best_llk;

   int chance_findings = 0;
   if (permutations)
      {
      printf("Haplotyping Random Permutations of the Data\n"
             "===========================================\n\n");

      for (int i = 0; i < permutations; i++)
         {
         SampleFilters();

         // Cases
         MaximizeLikelihood(info, caseFilter);
         double lk_A = likelihood.best_llk;

         // Controls
         MaximizeLikelihood(info, controlFilter);
         double lk_B = likelihood.best_llk;

         if (lk_A + lk_B > lkCases + lkControls)
            chance_findings++;

         printf("Permutation %4d: llk(cases) = %8.3f, llk(controls) = %8.3f, llk(sum) = %8.3f\n",
                i + 1, lk_A, lk_B, lk_A + lk_B);
         }
      }

   printf("\n\n"
          "Summary of Results\n"
          "==================\n\n"
          "logLikelihood for Combined Sample:  %8.3f\n"
          "logLikelihood for Cases:            %8.3f\n"
          "logLikelihood for Controls:         %8.3f\n"
          "logLikelihood for Cases + Controls: %8.3f\n"
          "logLikelihood ratio:                %8.3f\n",
          lkCombined, lkCases, lkControls, lkCases + lkControls,
          lkCases + lkControls - lkCombined);

   if (permutations)
      printf("Permutations with higher ratio:   %6d/%d\n",
             chance_findings, permutations);
   }

 
