////////////////////////////////////////////////////////////////////// 
// fugue/HaploSet.cpp 
// (c) 2000-2003 Goncalo Abecasis
// 
// This file is distributed as part of the FUGUE source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile FUGUE.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Tuesday July 06, 2004
// 
 
#include "HaploSet.h"
#include "Error.h"
#include "StringArray.h"

#include <math.h>

int HaplotypeSet::maxHaploCount = 0;

HaplotypeSet::HaplotypeSet()
   {
   size = haploCount = graphCount = 0;
   graphs = NULL;
   weight = 0.0;
   }

HaplotypeSet::~HaplotypeSet()
   {
   if (graphs != NULL)
      {
      for (int i = 0; i < graphCount; i++)
         delete graphs[i];
      delete [] graphs;
      }
   }

void HaplotypeSet::Grow()
   {
   size = size == 0 ? 1 : size * 2;

   HaplotypeGraph ** new_graphs = new HaplotypeGraph * [size];

   for (int i = 0; i < graphCount; i++)
      new_graphs[i] = graphs[i];

   if (graphs != NULL) delete [] graphs;

   graphs = new_graphs;
   }

void HaplotypeSet::LoadFromFile(FILE * file)
   {
   String input;
   StringArray tokens;

   do {
      input.ReadLine(file);
      tokens.Clear();
      tokens.AddTokens(input, " \t\f,");
   } while (tokens.Length() == 0);

   if (tokens.Length() != 6)
      error("Problem header encountered:\n\n%s",
            (const char *) input);

   HaplotypeGraph * graph = new HaplotypeGraph;

   int haplos = tokens[3];

   if (haploCount != 0 && haploCount != haplos)
      error("Expecting %d Haplotypes in Family %d\n",
            haploCount, (const char *) setid);

   haploCount = haplos;
   if (haplos > maxHaploCount)
      maxHaploCount = haploCount;

   graph->LoadFromFile(file, haplos);
   weight += graph->weight = tokens[5];

   if (graphCount >= size) Grow();

   graphs[graphCount++] = graph;
   }

void HaplotypeSet::StandardizeWeights()
   {
   if (weight != 1.0)
      {
      for (int i = 0; i < graphCount; i++)
         graphs[i]->weight /= weight;
      weight = 1.0;
      }
   }


double HaplotypeSet::Complexity()
   {
   double c = 0;

   for (int i = 0; i < graphCount; i++)
      c += graphs[i]->Complexity();

   return c;
   }

void HaplotypeSet::UpdateAlleleCounts()
   {
   for (int i = 0; i < graphCount; i++)
      graphs[i]->UpdateAlleleCounts();
   }

void HaplotypeSets::LoadFromFile(FILE * file)
   {
   ::String input;
   StringArray tokens;

   do {
      if (feof(file)) return;

      input.ReadLine(file);
      tokens.Clear();
      tokens.AddTokens(input);
   } while (tokens.Length() == 0);

   do {
      if (feof(file)) return;

      input.ReadLine(file);
      tokens.Clear();
      tokens.AddTokens(input);
   } while (tokens.Length() == 0);

   SetMarkerCount(tokens.Length() - 1);

   rewind(file);
   fpos_t pos;

   while (!feof(file))
      {
      fgetpos(file, &pos);

      do {
         if (feof(file)) return;

         input.ReadLine(file);
         tokens.Clear();
         tokens.AddTokens(input, " \t\r\n,");
      } while (tokens.Length() == 0);

      if (tokens.Length() != 6)
         error("Problem header encountered:\n\n%s",
               (const char *) input);

      HaplotypeSet * set = FindSet(tokens[1]);

      fsetpos(file, &pos);

      set->setid = tokens[1];
      set->LoadFromFile(file);
      }
   }

void * HaplotypeSets::create_set()
   {
   return new HaplotypeSet;
   }

HaplotypeSets::~HaplotypeSets()
   {
   for (int i = 0; i < Length(); i++)
      delete (HaplotypeSet *) objects[i];
   }

void HaplotypeSets::ShowCounts()
   {
   int totalHaplotypes = 0;
   int totalSets = 0;
   int knownHaplotypes = 0;
   int knownSets = 0;

   for (int i = Length() - 1; i >= 0; i--)
      {
      HaplotypeSet * set = (HaplotypeSet *) objects[i];

      if (set->Complexity() == 1.0)
         {
         knownHaplotypes += set->haploCount;
         knownSets++;
         }

      totalHaplotypes += set->haploCount;
      totalSets++;
      }

   printf("Total: %d Haplotypes in %d Sets\n"
          "Known: %d Haplotypes in %d Sets [UNDERESTIMATE]\n\n",
          totalHaplotypes, totalSets,
          knownHaplotypes, knownSets);
   }

void HaplotypeSets::Filter(int maxBits)
   {
   int totalHaplotypes = 0;
   int totalSets = 0;
   int knownHaplotypes = 0;
   int knownSets = 0;

   printf("Filtering data...\n");
   for (int i = Length() - 1; i >= 0; i--)
      {
      HaplotypeSet * set = (HaplotypeSet *) objects[i];

      double c = set->Complexity();
      double bits = log(c) / log(2.0);

      if (bits > maxBits)
         {
         printf("DISCARDING family %s (%.1f bits)\n",
               (const char *) set->setid, bits);

         Delete(i);
         delete set;
         continue;
         }

      if (bits == 0.0)
         {
         knownHaplotypes += set->haploCount;
         knownSets++;
         }

      totalHaplotypes += set->haploCount;
      totalSets++;
      }

   printf("\nTotal: %d Haplotypes in %d Sets\n"
          "Known: %d Haplotypes in %d Sets [UNDERESTIMATE]\n\n",
          totalHaplotypes, totalSets,
          knownHaplotypes, knownSets);
   }

void HaplotypeSets::UpdateAlleleCounts()
   {
   for (int i = 0; i < count; i++)
      ((HaplotypeSet *) objects[i])->UpdateAlleleCounts();
   }
 
