////////////////////////////////////////////////////////////////////// 
// fugue/Main.cpp 
// (c) 2000-2003 Goncalo Abecasis
// 
// This file is distributed as part of the FUGUE source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile FUGUE.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Tuesday July 06, 2004
// 
 
#include "HaploSet.h"
#include "Parameters.h"
#include "Error.h"
#include "Likelihood.h"
#include "SparseLikelihood.h"

void ShowHeader();
void ReadParameters(int argc, char ** argv);
void LoadData();

// Input haplotype data
HaplotypeSets sets;

// Name for input file
String filename("merlin.hap");

// Maximum complexity for haplotype sets in analysis
int maxBits = 16;

// Number of restarts using random haplotype frequencies
int restarts = 0;

// Print additional information for incremental builds
bool additional = false;

// Convergence threshold
double tol = 1e-6;

// Display threshold
double minfreq = 0.0001;

// Print additional information for two marker haplotypes
bool two_marker = false;

// Divide And Conquer Approximation
bool useApproximation = false;

int main(int argc, char ** argv)
   {
   ShowHeader();

   ReadParameters(argc, argv);
   LoadData();

   sets.UpdateAlleleCounts();

   if (!useApproximation)
      {
      sets.Filter(maxBits);

      Likelihood likelihood;

      likelihood.tol = tol;
      likelihood.Initialize();
      likelihood.EM(&sets);

      for (int i = 0; i < restarts; i++)
         likelihood.RandomEM(&sets);

      likelihood.Print(minfreq);

      if (additional) likelihood.PrintExtras();
      if (additional) likelihood.SaveSummary();
      if (two_marker) likelihood.PrintTwoMarker();
      }
   else
      {
      sets.ShowCounts();

      SparseLikelihood sparse;

      sparse.tol = tol;
      sparse.Initialize();
      sparse.EM(&sets);

      sparse.Print(minfreq);

      if (additional) sparse.PrintExtras(1e-7);
      }

   return 0;
   }

void ShowHeader()
   {
   printf("FUGUE - Frequency Using Graphs\n"
          "(c) 2001 Goncalo Abecasis\n");
   }

void ReadParameters(int argc, char ** argv)
   {
   ParameterList pl;
   pl.Add(new StringParameter('f', "Input File", filename));
   pl.Add(new IntParameter('b', "Max Bits", maxBits));
   pl.Add(new IntParameter('r', "Restarts", restarts));
   pl.Add(new DoubleParameter('c', "Convergence Threshold", tol));
   pl.Add(new DoubleParameter('t', "Display Threshold", minfreq));
   pl.Add(new HiddenSwitch('i', "Additional Information", additional));
   pl.Add(new HiddenSwitch('2', "Two Marker LD", two_marker));
   pl.Add(new SwitchParameter('a', "Divide-And-Conquer", useApproximation));
   pl.Read(argc, argv);
   pl.Status();
   }

void LoadData()
   {
   FILE * input = fopen(filename, "rb");
   if (input == NULL)
      error("Error opening file %s\n", (const char *) filename);

   sets.LoadFromFile(input);
   fclose(input);
   }
 
